#!/usr/bin/python3

import argparse
import gitlab
import logging
import sys
from config import config

RETRY_NONE = 0
RETRY_PIPELINE = 1
RETRY_JOBS = 2
RETRY_RETRIGGER = 3

def expand_projects(config):
    projects_to_check = set(config['projects'])

    for projet in projects_to_check:
        print(f'Expanding {projet} to:')

        if projet.endswith("/*"):
            projet_path = projet.replace('/*', '')
            projet_group = gitlab_check.gitlab.groups.get(projet_path, lazy=True)
            list_subprojects = projet_group.projects.list(get_all=True)

            for subproject in list_subprojects:
               subproject_id = gitlab_check.gitlab.projects.get(subproject.id)
               subproject_branches = subproject_id.branches.list(get_all=True)
               subproject_branches = [projet.name for projet in subproject_branches]

               if set(config['ref']) & set(subproject_branches):
                  config['projects'].append(subproject_id.path_with_namespace)
                  print(f'  {subproject_id.path_with_namespace}')

            config['projects'].remove(projet)

    return config

class GitlabCheck():
    def __init__(self, url = None, token = None, instance = None):
        if url:
            self.gitlab = gitlab.Gitlab(url, token)
        elif instance:
            self.gitlab = gitlab.Gitlab.from_config(config['instance'])
        elif config['url']:
            self.gitlab = gitlab.Gitlab(config['url'], config['token'])
        elif config['instance']:
            self.gitlab = gitlab.Gitlab.from_config(config['instance'])
        else:
            self.gitlab = gitlab.Gitlab.from_config()

        self.gitlab.auth()

    def fetch_pipelines(self, project_name, fetch_all = False):
        project = self.gitlab.projects.get(project_name)
        pipelines = project.pipelines.list(get_all=fetch_all)

        return pipelines

    def fetch_branches(self, project_name):
        project = self.gitlab.projects.get(project_name)
        branches = project.branches.list(get_all=True)

        return branches

    def existing_branches(self, project_name, pipelines):
        branches = self.fetch_branches(project_name)
        branch_names = [x.name for x in branches]
        pipelines = [x for x in pipelines if x.ref in branch_names]

        return pipelines

    def newest_pipelines(self, pipelines):
        pipelines_cache = {}
        for p in pipelines:
            if p.ref in pipelines_cache.keys():
                if pipelines_cache[p.ref].id < p.id:
                    pipelines_cache[p.ref] = p
            else:
                pipelines_cache[p.ref] = p

        return pipelines_cache.values()

    def filter_pipelines(self, pipelines, ref = None, status = None):
        ref = ref or config['ref']
        if ref:
            pipelines = [p for p in pipelines if p.ref in ref]
        status = status or config['status']
        if status and status != '-':
            pipelines = [p for p in pipelines if p.status == status]

        return pipelines

    def print_pipelines(self, project_name, pipelines, verbose = 1):
        for p in pipelines:
            if verbose == 1:
                print(project_name, p.id, p.ref, p.created_at, p.status)
            elif verbose == 2:
                print(project_name, p.id, p.ref, p.created_at, p.status, p.web_url)
            else:
                print(p)

    def retry_pipelines(self, project_name, pipelines):
        for p in pipelines:
            print(f'Retrying pipelines for {project_name} {p.ref}')
            p.retry()

    def retry_jobs(self, project_name, pipelines, fetch_all = False):
        def is_in_cache(jobs_cache, job):
            for j in jobs_cache:
                if  j.name  == job.name and j.id > job.id:
                    return True
                return False

        for p in pipelines:
            project = self.gitlab.projects.get(project_name)
            jobs = project.jobs.list(get_all=fetch_all)
            jobs_cache = []
            for j in jobs:
                if j.ref != p.ref:
                    continue
                if is_in_cache(jobs_cache, j):
                    continue
                jobs_cache.append(j)
                if j.status == 'failed':
                    print(f'Retrying job {j.name} {j.ref}')
                    j.retry()

    def retrigger_pipeline(self, project_name, pipelines):
        for p in pipelines:
            project = self.gitlab.projects.get(project_name)
            pipeline = project.pipelines.create({'ref': p.ref})
            print(f'Pipeline {pipeline}')

    def check_pipelines(self, all, ref, status, fetch_all = False, retry = False, verbose = 1):
        pipeline_count = 0
        for p in config['projects']:
            pipelines = self.fetch_pipelines(p, fetch_all)
            if not all:
                pipelines = self.existing_branches(p, pipelines)
                pipelines = self.newest_pipelines(pipelines)
            pipelines = self.filter_pipelines(pipelines, ref, status)
            pipeline_count += len(pipelines)
            self.print_pipelines(p, pipelines, verbose)
            if retry == RETRY_PIPELINE:
                self.retry_pipelines(p, pipelines)
            if retry == RETRY_JOBS:
                self.retry_jobs(p, pipelines, fetch_all)
            elif retry == RETRY_RETRIGGER:
                self.retrigger_pipeline(p, pipelines)

        return pipeline_count

if __name__ == '__main__':
    logging.basicConfig(level = logging.INFO)

    action_choices = [
        'check-pipelines',
    ]

    available_actions = ' '.join(action_choices)

    parser = argparse.ArgumentParser()
    parser.add_argument("-a", "--all", action="store_true", help="not only newest objects")
    parser.add_argument("-c", "--config", default="config.yaml", help="configuration file")
    parser.add_argument("-e", "--error", action="store_true", help="trigger error in case of output")
    parser.add_argument("-f", "--fetch_all", action="store_true", help="fetch all objects")
    parser.add_argument("-i", "--instance", help="Gitlab instance")
    parser.add_argument("-r", "--ref", action='append', help="filter by ref")
    parser.add_argument("-s", "--status", default=None, help="filter by status")
    parser.add_argument("-t", "--token", default="", help="Gitlab token")
    parser.add_argument("-u", "--url", help="Gitlab URL")
    parser.add_argument('-v', '--verbose', type=int, default=1, help='verbose')
    parser.add_argument("-x", "--retry", type=int, default=RETRY_NONE, help="retry pipeline: 0 none, 1 pipeline, 2 jobs, 3 retrigger")
    parser.add_argument('action', choices=action_choices, help="action to be executed, available actions: " + available_actions, metavar='action')

    args = parser.parse_args()

    config.load_config(args)

    gitlab_check = GitlabCheck(args.url, args.token, args.instance)

    config = expand_projects(config)

    if args.action == 'check-pipelines':
        pipeline_count = gitlab_check.check_pipelines(args.all, args.ref, args.status, args.fetch_all, args.retry, args.verbose)
        if args.error and pipeline_count:
            print(f'There are {pipeline_count} pipelines with failure')
            sys.exit(1)
    else:
        print('Invalid action')
